﻿namespace U2UConsult.DockOfTheBay.LinqToSpatialSample
{
    using System;
    using System.Data.Linq;
    using System.Data.SqlTypes;
    using System.Linq;
    using System.Text;
    using Microsoft.SqlServer.Types;

    /// <summary>
    /// Demonstrates the use of Spatial data types in Linq to SQL.
    /// </summary>
    public class Program
    {
        public static void Main(string[] args)
        {
            using (EuropeDataContext db = new EuropeDataContext())
            {
                // Uncomment to inspect the queries:
                // db.Log = Console.Out;

                Console.WriteLine("\n ----------------------------------------------");
                Console.ForegroundColor = ConsoleColor.Green;
                Console.Write("     U2U Consult");
                Console.ForegroundColor = ConsoleColor.White;
                Console.WriteLine(" Linq to SQL Spatial Sample");
                Console.WriteLine(" ----------------------------------------------\n");

                ////
                // Selecting from the View ('Shape' is a Binary)
                ////

                Console.ForegroundColor = ConsoleColor.Cyan;
                Console.WriteLine(" Hint 1: Select from a view.");
                Console.ForegroundColor = ConsoleColor.White;

                Binary belgiumb = (from c in db.vw_Countries
                                   where c.CountryID == 4
                                   select c.Shape).FirstOrDefault();
                SqlGeography geo = new SqlGeography();

                // Potential NullReference (if query has no result)
                geo.Read(new System.IO.BinaryReader(new System.IO.MemoryStream(belgiumb.ToArray())));

                ////
                // For the record: the area is not accurate since it's calculated on 
                // a highly reduce shape.
                ////

                Console.WriteLine(" The area of Belgium is {0} m²", geo.STArea().Value);

                Console.ForegroundColor = ConsoleColor.Cyan;
                Console.WriteLine("\n Hint 2: Wrap conversions in extension methods.");
                Console.ForegroundColor = ConsoleColor.White;

                SqlGeography b = (from c in db.vw_Countries
                                  where c.CountryID == 4
                                  select c.Shape).FirstOrDefault().AsGeography();

                Console.WriteLine(" The area of Belgium is {0} m²", b.STArea().Value);

                ////
                // Selecting from the Table ('Shape' is a Binary)
                ////

                Console.ForegroundColor = ConsoleColor.Cyan;
                Console.WriteLine("\n Hack 1: Override the data source mapping.");
                Console.ForegroundColor = ConsoleColor.White;

                string belgiumShape = (from c in db.Countries
                                       where c.CountryID == 4
                                       select c.Shape.ToString()).FirstOrDefault();
                SqlGeography belgium = SqlGeography.Parse(belgiumShape);

                // Or in 1 sentence:
                belgium =
                    SqlGeography.Parse((from c in db.Countries
                                        where c.CountryID == 4
                                        select c.Shape.ToString()).FirstOrDefault());

                Console.WriteLine(" The area of Belgium is {0} m²", belgium.STArea().Value);

                ////
                // Selecting from the Table ('Shape' is a SqlGeography)
                ////

                Console.ForegroundColor = ConsoleColor.Cyan;
                Console.WriteLine("\n Hack 2: Override the data type mapping.");
                Console.ForegroundColor = ConsoleColor.White;

                // Using a spatial type in the Select-clause

                // InvalidCastException 
                // Builds a query, sends sends it to SQL Server, but fails to deserialize the result:
                //
                // belgium = (from c in db.Countries2
                //            where c.CountryID == 4
                //            select c.Shape).FirstOrDefault();

                // Console.WriteLine(belgium.STArea().Value);

                // Workaround -> not much gain compared to Countries
                belgium =
                    SqlGeography.Parse((from c in db.Countries2
                                        where c.CountryID == 4
                                        select c.Shape.ToString()).FirstOrDefault());

                Console.WriteLine(" The area of Belgium is {0} m²", belgium.STArea().Value);

                // Using a spatial type in the Where-clause

                // NotSupportedException (unable to build good query):
                //
                // var query = (from c in db.Countries2
                //              where c.Shape.STIntersects(SqlGeography.Parse(belgium.ToString())).Value
                //              select c.CountryName).First();
                // var query = (from c in db.Countries2
                //              where (c.Shape.STArea() > 10).Value
                //              select c.CountryName).FirstOrDefault();

                ////
                // Packaging spatial method calls in scalar functions
                ////

                Console.ForegroundColor = ConsoleColor.Cyan;
                Console.WriteLine("\n Hint 3: Define scalar functions.");
                Console.ForegroundColor = ConsoleColor.White;

                // Clustered Index Scan - Always
                var query = from c in db.Countries
                            where db.Intersects(c.Shape.ToString(), belgium.ToString()) == 1
                            select c.CountryName;

                foreach (string country in query)
                {
                    Console.WriteLine(" " + country);
                }

                ////
                // Packaging spatial method calls in table valued functions
                ////

                Console.ForegroundColor = ConsoleColor.Cyan;
                Console.WriteLine("\n Hint 4: Define table valued functions.");
                Console.ForegroundColor = ConsoleColor.White;

                // Query Plan cost is 5x worse ??
                // But in reality CPU and Duration 5x better !!
                var query2 = from c in db.Intersectors(belgium.ToString())
                             select c.CountryName;

                foreach (string country in query2)
                {
                    Console.WriteLine(" " + country);
                }
            }

            // The end
            Console.WriteLine("\n ----------------------------------------------");
            Console.WriteLine("           http://www.u2uconsult.com");
            Console.WriteLine(" ----------------------------------------------");

            Console.ReadKey();
        }
    }
}
